import os
import pandas as pd

def find_owner(uid):
    from pwd import getpwuid
    return getpwuid(uid).pw_name

def unixtodate(ts):
    from datetime import datetime
    return datetime.fromtimestamp(ts) #.strftime('%Y-%m-%d %H:%M:%S')

class FileEntry():
    import os

    def __init__(self, file:os.DirEntry, rootdir = None):
        self.name = file.name
        self.fullname = file.path

        self.path = file.path[:-1 * len(file.name)]

        f = file.name.rfind('.')
        self.extension = '' if f == -1 else file.name[f + 1 :]
        self.basename = self.name if f == -1 else file.name[: f]
        self.isdir = file.is_dir()
        self.isfile = file.is_file()
        self.issymlink = file.is_symlink()
        stat = file.stat()
        self.size = stat.st_size
        self.created = unixtodate(stat.st_ctime)
        self.modified = unixtodate(stat.st_mtime)
        self.changed = unixtodate(stat.st_ctime)
        self.owner = find_owner(stat.st_uid)
        if rootdir:
            if rootdir[-1] == '/':
                rootdir = rootdir[:-1]
            self.parent = self.path[len(rootdir):]
            self.level = self.parent.count('/') # + 0 if len(x) > 0 and x[-1] == '/' else 1
        else:
            self.level = 0
            self.parent = '.'

    def __repr__(self) -> str: 
        return f'Name = "{self.name}" Path = "{self.path}" Size = {self.size} Owner = {self.owner} Created = {self.created} Modified = {self.modified} Changed = {self.changed}'
    
def scantree(path, follow_symlinks = False, recursive = True):
    for entry in os.scandir(path):
        if recursive and entry.is_dir(follow_symlinks = follow_symlinks):
            yield entry
            yield from scantree(entry.path)
        else:
            yield entry

 

# def get_table_names(sql):
#     import re
#     x = re.split('from |FROM |From |join |JOIN |Join ', sql)
#     print('******')
#     tables = [t[:f'{t} .'.index(' ')] for t in x][1:]
#     return tables


def sqldir(sql):
    from sqldf import sqldf, get_table_names
    from os.path import expanduser
    tablenames = get_table_names(sql)

    if len(tablenames) >= 1:
        sql0 = sql
        dirs = dict()
        cnt = 1
        for t in tablenames: 
            sql0 = sql0.replace(t, f'd{cnt}')
            scan = [FileEntry(f, rootdir = expanduser(t)) for f in scantree(expanduser(tablenames[cnt - 1]))]
            scandf = pd.DataFrame((o.__dict__ for o in scan))
            dirs[f'd{cnt}'] = scandf
            cnt += 1

        r = sqldf(sql0, **dirs)
        return r

if __name__ == '__main__':
    import sys
    args = sys.argv
    print(args)
    if len(args) > 1:
        sql = args[1]
        print(sql)
    else:
        sql = '''select name, path, size, level, parent from ~/Downloads/dir1/ as t1'''
        #sql = '''select name, path, size, level, isfile from ~/Downloads/dir1 as t1 where level = 1'''
        sql = '''select name, path, size, level, parent from ~/Downloads/dir1/ as t1 where name like "file%"'''

 
    # sql = '''select * from /Users/joey/Downloads/dir1 as t1 
    # join /Users/joey/Downloads/dir2 as t2 on t1.name = t2.name'''

    # sql = '''select t1.name as name1, t2.name as name2 from /Users/joey/Downloads/dir1 as t1 
    # join /Users/joey/Downloads/dir2 as t2 on t1.name = t2.name'''

    print(sqldir(sql))
    import os.path as p
    f = '~/Downloads/dir1/file1.csv'
    f1 = p.expanduser(f)
    print(f'{p.basename(f)=} {p.abspath(f)}= {p.dirname(f)=} {p.relpath(f, "/Users/Downloads/")=}')


'''
Written by Joseph Gagliardo joey@me.com on 2021-06-05
This module is meant to offer a more flexible version of what PandaSQL does, by providing support for Python list and dict
and the ability to use Python UDF within the SQL. 

Furthermore, you can use SQL UPDATE, INSERT, DELETE commands to modify the DataFrame object and return a new DataFrame with those results.

You can also CREATE TABLE and populate it in SQL and combine it with the source from a DataFrame or collection
'''

def get_table_names(sql):
    import re
    x = re.split('from |FROM |From |join |JOIN |Join ', sql)
    tables = [t[:f'{t.strip()} '.index(' ')] for t in x][1:]
    return tables

def sqldf(sql:str, *, index:bool = False, output:str = 'dataframe', **params):
    '''
    sql : str
         The SQL command or commands you want to run. 
         You can pass in multiple statements separated by a ; in order to do UPDATE, INSERT, DELETE, but the last query must be a SELECT without a ;
    
    index : bool (Default: False)
         When a DataFrame is pushed into the memory table, should the index col be included

    output : str {'dataframe', 'dict', 'list', 'series', 'split', 'records', 'index'}
        Determines the type of the values of the dictionary.
    
        - 'dict' (default) : dict like {column -> {index -> value}}
        - 'list' : dict like {column -> [values]}
        - 'series' : dict like {column -> Series(values)}
        - 'split' : dict like {'index' -> [index], 'columns' -> [columns], 'data' -> [values]}
        - 'records' : list like [{column -> value}, ... , {column -> value}]

    params : dict
        KV parameters to pass in Python functions to the memory database, 
        or to specify specific tables with aliases rather than the default of searching for FROM and JOIN
        key is a string which is the function name SQL will use, 
        value is a pointer to the function or table source

        There is no real error checking so make sure any list or dicts you pass in to automatically convert to a DataFrame
        are in the correct format and that you don't send any bad SQL. This module relies on the errors thrown by the 
        functions it calls.
    '''

    from pandas import read_sql_query, DataFrame
    import sqlite3
    import types 
    from inspect import signature, stack

    # get globals from the caller level in order to find objects that refer to virtual tables in the SQL statement
    env = stack()[1][0].f_globals


    with sqlite3.connect(':memory:') as cn:
        addedtables = set()
        # go through the list of KV parameters to push any UDF's or named tables into the memory database
        for k, v in params.items():
            if type(v) in (types.FunctionType, types.LambdaType):
                # If it's a function add it
                cn.create_function(k, len(signature(v).parameters), v)
            elif isinstance(v, DataFrame):
                # if it's a DataFrame add it as is
                addedtables.add(k)
                v.to_sql(k, cn, index = index)
            elif isinstance(v, dict) or isinstance(v, list):
                # if it's a list of dict automatically convert it to a DataFrame
                addedtables.add(k)
                df = DataFrame(v)
                df.to_sql(k, cn, index = index)

        # search through the environment to see if any tables mentioned in FROM or JOIN clause 
        # that are not specified in the params and push them into the memory database
        # the table can be a DataFrame or a list or dict that is convertible to a DataFrame
        for k in get_table_names(sql):
            if k in env and k not in addedtables:
                o = env[k]
                if isinstance(o, DataFrame):
                    o.to_sql(k, cn, index = index)
                elif isinstance(o, dict) or isinstance(o, list):
                    df = DataFrame(o)
                    df.to_sql(k, cn, index = index)

        # you can pass in multiple statements separated by a ; in order to do CREATE TABLE, UPDATE, INSERT, DELETE, 
        # but the last query must be a SELECT to return a result
        if ';' in sql:
            commands = [x.strip() for x in sql.split(';')]
            for c in commands[:-1]:
                cn.execute(c)
            r = read_sql_query(commands[-1], cn)
        else:
            r = read_sql_query(sql, cn)

        # if the output flag is passed it calls the DataFrame to_dict with that option
        if output and output.lower() != 'dataframe':
            return r.to_dict(output)
        
        return r



import pandas as pd

# if running locally during builds before uploading to pypi
#from sqldf import sqldf

# if running to test the downloaded package from pypi
from bettersql import sqldf

def reverse(x):
    # sample function with one parameter
    if x is not None:
        return x[::-1]

def twofun(x, y):
    # sample function with two parameters
    return x * y

# DataFrame source
names = pd.DataFrame({'id':[1, 2, 3], 'name':['Alpha', 'Beta', 'Gamma'], 'category':[1, 2, 2]})    

# Dict source of lists
categories = {'id':[1, 2, 3], 'name':['One', 'Two', 'Three']}

# List of dicts as source
languages = [{'id':1, 'language':'english'}, {'id':2, 'language':'italian'}]

sql = '''
SELECT n.id, n.name, n.category, c.name as categoryname
    , reverse(n.name) as reverse, rev(c.name) as rev, twofun(n.name, n.id) as repeat 
FROM names AS n
LEFT JOIN categories AS c on n.category= c.id
'''

r = sqldf(sql, reverse = reverse, twofun = twofun, rev = lambda x : None if x is None else x[::-1]) 
print('-' * 30)
print(r)

# alternatively you can pass in the table names as named parameters like you do for function names, this allows you to alias them to a different name
sql = '''
SELECT n.id, n.name, n.category, c.name as categoryname
    , reverse(n.name) as reverse, rev(c.name) as rev, twofun(n.name, n.id) as repeat 
FROM names2 AS n
LEFT JOIN categories AS c on n.category= c.id
'''
r = sqldf(sql, reverse = reverse, twofun = twofun
        , rev = lambda x : None if x is None else x[::-1], names2 = names, categories = categories) 
print('-' * 30)
print(r)


# You could mix the two modes also, to specify some tables and have the others automatically read
r = sqldf(sql, reverse = reverse, twofun = twofun, rev = lambda x : None if x is None else x[::-1], names2 = names) 
print('-' * 30)
print(r)

sql = '''
update names set name = 'Omega' where id = 1;
insert into names values(4, 'Pi', 4);
SELECT n.id, n.name, n.category, c.name as categoryname
    , reverse(n.name) as reverse, rev(c.name) as rev, l.language
FROM names AS n
LEFT JOIN categories AS c on n.category = c.id
LEFT JOIN languages as l on n.category = l.id
'''

r = sqldf(sql, reverse = reverse, rev=lambda x : None if x is None else x[::-1])
print('-' * 30)
print(r)

# Could also create a table in the sql connection and use it for the SELECT query
sql = '''

create table other(id int, name string);
insert into other select 1, 'other' 
union all select 2, 'thing';
SELECT n.id, n.name, n.category, o.name as othername
FROM names AS n
LEFT JOIN other as o on n.category = o.id
'''

print('-' * 30)
print('DataFrame output')
r = sqldf(sql)
print(r)

print('-' * 30)
print('DataFrame output')
r = sqldf(sql, output = 'dataframe')
print(r)

# can also output the result into modes supported by to_dict method of DataFrame
print('-' * 30)
print('dict output')
r = sqldf(sql, output = 'dict')
print(r)

print('-' * 30)
r = sqldf(sql, output = 'list')
print('list output')
print(r)

print('-' * 30)
r = sqldf(sql, output = 'records')
print('records output')
print(r)




from setuptools import setup, find_packages

VERSION = '1.3.1' 
DESCRIPTION = 'A package to use SQL with Pandas DataFrames, list and dict.'
LONG_DESCRIPTION = 'A more feature rich alternative to Pandasql that allows for custom UDF, as well as lit and dict as source tables and return types'

from os import path
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md')) as f:
    long_description = f.read()

# Setting up
setup(
       # the name must match the folder name 'verysimplemodule'
        name="bettersql", 
        version=VERSION,
        author="Joseph Gagliardo",
        author_email="joey@me.com",
        description=DESCRIPTION,
        long_description=long_description,
        long_description_content_type='text/markdown',
        install_requires=['pandas'], 
        keywords=['python', 'sql', 'pandas', 'udf'],
        packages=find_packages(),
        classifiers= [
            "Development Status :: 5 - Production/Stable",
            "Intended Audience :: Education",
            "Intended Audience :: Developers",
            "Programming Language :: Python :: 3",
            "Operating System :: MacOS :: MacOS X",
            "Operating System :: POSIX :: Linux",
            "License :: OSI Approved :: Apache Software License"
        ]
        , project_urls = {"Home Page" : "https://github.com/joegagliardo/bettersql"}
)

